
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include "radeon.h"
#include "radeon_drm.h"
#include "radeon_bufmgr_gem.h"

Bool
radeon_bind_memory(ScrnInfoPtr pScrn, struct radeon_memory *mem)
{
	RADEONInfoPtr info = RADEONPTR(pScrn);	
  
	if (mem == NULL || mem->bound)
		return TRUE;

	if (!info->drm_mm)
		return FALSE;

	if (mem->kernel_bo_handle) {
		struct drm_radeon_gem_pin pin;

		int ret;

    		if (mem->pool == RADEON_POOL_VRAM)
		    pin.pin_domain = RADEON_GEM_DOMAIN_VRAM;
    		else
      		    pin.pin_domain = RADEON_GEM_DOMAIN_GTT;
		pin.handle = mem->kernel_bo_handle;
		pin.alignment = mem->alignment;

		ret = ioctl(info->dri->drmFD, DRM_IOCTL_RADEON_GEM_PIN, &pin);
		if (ret != 0) {
			xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
				   "Failed to pin %s: %s\n", mem->name, strerror(errno));
			return FALSE;
		}

		mem->bound = TRUE;
		mem->offset = pin.offset;
		//		ErrorF("pin returned 0x%llx\n", pin.offset);
		mem->end = mem->offset + mem->size;
		return TRUE;
	}
	return FALSE;
}

static Bool
radeon_unbind_memory(ScrnInfoPtr pScrn, struct radeon_memory *mem)
{
	RADEONInfoPtr info = RADEONPTR(pScrn);
	int ret;

	if (mem == NULL || !mem->bound)
		return TRUE;

	if (!info->drm_mm)
		return FALSE;


	if (mem->kernel_bo_handle) {
		struct drm_radeon_gem_unpin unpin;

		unpin.handle = mem->kernel_bo_handle;
		ret = ioctl(info->dri->drmFD, DRM_IOCTL_RADEON_GEM_UNPIN, &unpin);

		if (ret == 0) {
			mem->bound = FALSE;
			mem->offset = -1;
			mem->end = -1;
			return TRUE;
		} else {
			return FALSE;
		}
	}
	return FALSE;
}

void radeon_free_memory(ScrnInfoPtr pScrn, struct radeon_memory *mem)
{
	RADEONInfoPtr info = RADEONPTR(pScrn);

	if (mem == NULL)
		return;

	if (mem->map)
	    radeon_unmap_memory(pScrn, mem);
	    
	radeon_unbind_memory(pScrn, mem);

	if (mem->kernel_bo_handle) {
		struct drm_gem_close close;

		close.handle = mem->kernel_bo_handle;
		ioctl(info->dri->drmFD, DRM_IOCTL_GEM_CLOSE, &close);
	}

	if (info->mm.bo_list[mem->pool] == mem) {
	  info->mm.bo_list[mem->pool] = mem->next;
	  if (mem->next)
	    mem->next->prev = NULL;
	} else {
	  if (mem->prev)
	    mem->prev->next = mem->next;
	  if (mem->next)
	    mem->next->prev = mem->prev;
	}
	xfree(mem->name);
	xfree(mem);
	return;
}

struct radeon_memory *radeon_allocate_memory(ScrnInfoPtr pScrn, int pool, int size, int alignment, Bool no_backing_store, char *name,
					     int static_alloc)
{
    RADEONInfoPtr  info   = RADEONPTR(pScrn);	
    struct drm_radeon_gem_create args;
    struct radeon_memory *mem;
    int ret;

    mem = xcalloc(1, sizeof(*mem));
    if (!mem)
	return NULL;
    
    mem->name = xstrdup(name); 
    if (!mem->name) {
	xfree(mem);
	return NULL;
    }

    mem->size = size;
    mem->pool = pool;
    mem->next = mem->prev = NULL;
    mem->vt_bind = static_alloc;
    args.size = size;
    args.alignment = alignment;
    if (pool == RADEON_POOL_VRAM)
      args.initial_domain = RADEON_GEM_DOMAIN_VRAM;
    else
      args.initial_domain = RADEON_GEM_DOMAIN_GTT;
    args.no_backing_store = no_backing_store;

    ret = drmCommandWriteRead(info->dri->drmFD, DRM_RADEON_GEM_CREATE, &args, sizeof(args));
    if (ret) {
	ErrorF("Failed to allocate %s\n", mem->name);
	xfree(mem);
	return NULL;
    }

    mem->kernel_bo_handle = args.handle;
    //    xf86DrvMsg(pScrn->scrnIndex, X_INFO,	
    //	       "%s allocated %d with handle %x\n", mem->name, mem->size, mem->kernel_bo_handle);

    /* add to VRAM linked list for now */

    mem->prev = NULL;
    mem->next = info->mm.bo_list[pool];
    if (info->mm.bo_list[pool] != NULL)
	info->mm.bo_list[pool]->prev = mem;
    info->mm.bo_list[pool] = mem;
    return mem;
}

Bool radeon_bind_all_memory(ScrnInfoPtr pScrn)
{
    RADEONInfoPtr  info   = RADEONPTR(pScrn);	
    struct radeon_memory *mem;
    int i;

    for (i = 0; i < 2; i++) {
	for (mem = info->mm.bo_list[i]; mem != NULL;
	     mem = mem->next) {
	    if (mem->vt_bind)
		if (!radeon_bind_memory(pScrn, mem)) {
		    FatalError("Couldn't bind %s\n", mem->name);
		}
	}
    }
    return TRUE;
}
	    
Bool radeon_unbind_all_memory(ScrnInfoPtr pScrn)
{
    RADEONInfoPtr  info   = RADEONPTR(pScrn);	
    struct radeon_memory *mem;
    int i;

    for (i = 0; i < 2; i++) {
	for (mem = info->mm.bo_list[i]; mem != NULL;
	     mem = mem->next) {
	    if (mem->vt_bind)
		radeon_unbind_memory(pScrn, mem);
	}
    }
    return TRUE;
}

Bool radeon_free_all_memory(ScrnInfoPtr pScrn)
{
    RADEONInfoPtr  info   = RADEONPTR(pScrn);	
    struct radeon_memory *mem, *tmp;
    int i;

    for (i = 0; i < 2; i++) {
	
	for (mem = info->mm.bo_list[i]; mem != NULL; ) {
	    tmp = mem->next;
	    radeon_free_memory(pScrn, mem);
	    mem = tmp;
	}
    }
    return TRUE;
}

int radeon_map_memory(ScrnInfoPtr pScrn, struct radeon_memory *mem)
{
    struct drm_radeon_gem_mmap args;
    RADEONInfoPtr info = RADEONPTR(pScrn);
    int ret;

    assert(!mem->map);


    args.handle = mem->kernel_bo_handle;
    args.size = mem->size;
    ret = drmCommandWriteRead(info->dri->drmFD, DRM_RADEON_GEM_MMAP, &args, sizeof(args));

    if (!ret)
	mem->map = (void *)(unsigned long)args.addr_ptr;
    //    ErrorF("Mapped %s size %ld at %x %p\n", mem->name, mem->size, mem->offset, mem->map);
    return ret;
}

void radeon_unmap_memory(ScrnInfoPtr pScrn, struct radeon_memory *mem)
{
    assert(mem->map);
    if (mem->map) {
        munmap(mem->map, mem->size);
        mem->map = NULL;
    }
}

/* Okay radeon
   2D allocations 
   - Front buffer
   - cursors
   - EXA space

   3D related:
   - Backbuffer
   - Depth buffer
   - textures
*/


Bool radeon_setup_kernel_mem(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    RADEONInfoPtr info = RADEONPTR(pScrn);
    xf86CrtcConfigPtr   xf86_config = XF86_CRTC_CONFIG_PTR(pScrn);
    int cpp = info->CurrentLayout.pixel_bytes;
    int screen_size;
    int stride = pScrn->displayWidth * cpp;
    int total_size_bytes = 0, remain_size_bytes;
    int fb_size_bytes;
    int pagesize = 4096;
    
    screen_size = RADEON_ALIGN(pScrn->virtualY, 16) * stride;

    ErrorF("%d x %d x %d = %dK\n", pScrn->displayWidth, pScrn->virtualY, cpp, screen_size / 1024);

    {
	int cursor_size = 64 * 4 * 64;
	int c;

    	cursor_size = RADEON_ALIGN(cursor_size, pagesize);
	for (c = 0; c < xf86_config->num_crtc; c++) {
	    /* cursor objects */
	    info->mm.cursor[c] = radeon_allocate_memory(pScrn, RADEON_POOL_VRAM, cursor_size, 0, 1, "Cursor", 1);
	    if (!info->mm.cursor[c]) {
		return FALSE;
	    }

	    radeon_bind_memory(pScrn, info->mm.cursor[c]);

	    if (radeon_map_memory(pScrn, info->mm.cursor[c])) {
		ErrorF("Failed to map front buffer memory\n");
	    }
	    
	    if (!info->drm_mode_setting) {
		xf86CrtcPtr crtc = xf86_config->crtc[c];
		RADEONCrtcPrivatePtr radeon_crtc = crtc->driver_private;
		radeon_crtc->cursor = info->mm.cursor[c];
	    } else {
		drmmode_set_cursor(pScrn, &info->drmmode, c, (void *)info->mm.cursor[c]->map, info->mm.cursor[c]->kernel_bo_handle);
	    }
	    total_size_bytes += cursor_size;
	}
    }

    screen_size = RADEON_ALIGN(screen_size, pagesize);
    /* keep area front front buffer - but don't allocate it yet */
    total_size_bytes += screen_size;

    if (info->directRenderingEnabled) {
	info->dri->backPitch = pScrn->displayWidth;
	info->mm.back_buffer = radeon_allocate_memory(pScrn, RADEON_POOL_VRAM, screen_size, 0, 1, "Back Buffer", 1);
	if (!info->mm.back_buffer) {
	    return FALSE;
	}
	radeon_bind_memory(pScrn, info->mm.back_buffer);
	total_size_bytes += screen_size;
	
	info->dri->depthPitch = RADEON_ALIGN(pScrn->displayWidth, 32);
	{
	    int depthCpp = (info->dri->depthBits - 8) / 4;
	    int depth_size = RADEON_ALIGN(pScrn->virtualY, 16) * info->dri->depthPitch * depthCpp;
    	    depth_size = RADEON_ALIGN(depth_size, pagesize);
	    info->mm.depth_buffer = radeon_allocate_memory(pScrn, RADEON_POOL_VRAM, depth_size, 0, 1, "Depth Buffer", 1);
	    if (!info->mm.depth_buffer) {
		return FALSE;
	    }
	    radeon_bind_memory(pScrn, info->mm.depth_buffer);
	    total_size_bytes += depth_size;
	}
    }

    /* work out from the mm size what the exa / tex sizes need to be */
    remain_size_bytes = info->mm.vram_size - total_size_bytes;

    if (info->dri->textureSize > 0)
    	info->dri->textureSize = (remain_size_bytes / 100) * info->dri->textureSize;
    else
    	info->dri->textureSize = remain_size_bytes / 2;

    info->dri->textureSize = RADEON_ALIGN(info->dri->textureSize, pagesize);

    remain_size_bytes -= info->dri->textureSize;

    ErrorF("texture size is %dK, exa is %dK\n", info->dri->textureSize / 1024, remain_size_bytes/1024);

    fb_size_bytes = screen_size;

    ErrorF("fb size is %dK %dK\n", fb_size_bytes / 1024, total_size_bytes / 1024);

    info->mm.front_buffer = radeon_allocate_memory(pScrn, RADEON_POOL_VRAM, fb_size_bytes, 0, 1, "Front Buffer + EXA", 1);
    if (!info->mm.front_buffer) {
	return FALSE;
    }

    radeon_bind_memory(pScrn, info->mm.front_buffer);
    if (radeon_map_memory(pScrn, info->mm.front_buffer)) {
	ErrorF("Failed to map front buffer memory\n");
    }
    info->dri->frontPitch = pScrn->displayWidth;

    if (info->directRenderingEnabled) {
	info->mm.texture_buffer = radeon_allocate_memory(pScrn, RADEON_POOL_VRAM, info->dri->textureSize, 0, 1, "Texture Buffer", 1);
	if (!info->mm.texture_buffer) {
	    return FALSE;
	}
	radeon_bind_memory(pScrn, info->mm.texture_buffer);
    }
	
    if (info->drm_mode_setting) {
	drmmode_set_fb(pScrn, &info->drmmode, pScrn->virtualX, RADEON_ALIGN(pScrn->virtualY, 16), stride, info->mm.front_buffer->kernel_bo_handle);
    }

    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Front buffer size: %dK at 0x%08x\n", info->mm.front_buffer->size/1024, info->mm.front_buffer->offset);
    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Back buffer size:  %dK at 0x%08x\n", info->mm.back_buffer->size/1024, info->mm.back_buffer->offset);
    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Depth buffer size: %dK at 0x%08x\n", info->mm.depth_buffer->size/1024, info->mm.depth_buffer->offset);
    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Texture size:      %dK at 0x%08x\n", info->mm.texture_buffer->size/1024, info->mm.texture_buffer->offset);
    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Remaining VRAM size (used for pixmaps): %dK\n", remain_size_bytes/1024);

    /* set the emit limit at 90% of VRAM */
    remain_size_bytes = (remain_size_bytes / 10) * 9;

    radeon_bufmgr_gem_set_limit(info->bufmgr, RADEON_GEM_DOMAIN_VRAM, remain_size_bytes);
    return TRUE;
}

Bool radeon_setup_gart_mem(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    RADEONInfoPtr info = RADEONPTR(pScrn);


    info->mm.gart_texture_buffer =
	radeon_allocate_memory(pScrn, RADEON_POOL_GART,
			       info->dri->gartTexMapSize,
			       0, 1, "GART texture buffers", 1);
    
    if (!info->mm.gart_texture_buffer) {
	return FALSE;
    }

    radeon_bind_memory(pScrn, info->mm.gart_texture_buffer);

    return TRUE;
}

uint32_t radeon_create_new_fb(ScrnInfoPtr pScrn, int width, int height, int *pitch)
{
    return 0;
}


dri_bo *radeon_create_rotate_bo(ScrnInfoPtr pScrn, int size)
{
	RADEONInfoPtr info = RADEONPTR(pScrn);
	dri_bo *bo;

	bo = dri_bo_alloc(info->bufmgr, "rotate", size, 0, 0);

	dri_bo_pin(bo, RADEON_GEM_DOMAIN_VRAM);
	return bo;
}

